from ..core import AbstractDistribution


class AbstractMultivariateDistribution(AbstractDistribution):
    pass


class ContinuousMultivariateDistribution(AbstractMultivariateDistribution):
    pass

        
